#include "images/PNGReader.h"

PNGReader::PNGReader()
    : path(""), width(0), height(0), channels(0), bitDepth(0), rowBytes(0),
      file(nullptr), read(nullptr), info(nullptr), rowPointers(nullptr) {
}

PNGReader::~PNGReader() {
    if(file != nullptr) {
        fclose(file);
    }
    if(rowPointers != nullptr) {
        png_free(read, rowPointers);
    }
    png_destroy_read_struct(&read, &info, nullptr);
}

Error PNGReader::load(const char* path) {
    if(file != nullptr) {
        Error error = {"file '"};
        error.message.append(path).append("' is alreaded loaded");
        return error;
    }
    PNGReader::path = path;
    file = fopen(path, "rb");
    if(file == nullptr) {
        Error error = {"file '"};
        error.message.append(path)
            .append("' cannot be read: ")
            .append(static_cast<const char*>(strerror(errno)));
        return error;
    }
    png_byte buffer[8];
    if(fread(buffer, sizeof(png_byte), 8, file) != 8) {
        Error error = {"cannot read signature of file '"};
        error.message.append(path).append("'");
        return error;
    }
    if(png_sig_cmp(buffer, 0, 8)) {
        Error error = {"file '"};
        error.message.append(path).append("' is not a png");
        return error;
    }
    read = png_create_read_struct(PNG_LIBPNG_VER_STRING, nullptr, nullptr,
                                  nullptr);
    if(read == nullptr) {
        return {"cannot create png read data structure"};
    }
    info = png_create_info_struct(read);
    if(info == nullptr) {
        return {"cannot create png info structure"};
    }
    if(setjmp(png_jmpbuf(read))) {
        Error error = {"png file '"};
        error.message.append(path).append("' has used error callback");
        return error;
    }
    png_init_io(read, file);
    png_set_sig_bytes(read, 8);
    png_read_info(read, info);

    png_set_expand(read);

    width = png_get_image_width(read, info);
    height = png_get_image_height(read, info);
    channels = png_get_channels(read, info);
    bitDepth = png_get_bit_depth(read, info);
    if(png_get_bit_depth(read, info) == 16) {
        png_set_strip_16(read);
    }
    rowBytes = png_get_rowbytes(read, info);

    if(channels < 1 || channels > 4) {
        Error error = {"'"};
        error.message.append(path)
            .append("' has unsupported number of channels: ")
            .append(channels);
        return error;
    } else if(width < 1 || width > 2048 || height < 1 || height > 2048) {
        Error error = {"width and height of '"};
        error.message.append(path).append("' are too big");
        return error;
    } else if(bitDepth != 8 && bitDepth != 16) {
        Error error = {"bit depth of '"};
        error.message.append(path).append("' is neither 8 or 16");
        return error;
    } else if(getBufferSize() != (rowBytes * height * 8 / bitDepth)) {
        Error error = {"'"};
        error.message.append(path).append("' needs an unexpected buffer size");
        return error;
    }
    return {};
}

int PNGReader::getWidth() const {
    return width;
}

int PNGReader::getHeight() const {
    return height;
}

int PNGReader::getChannels() const {
    return channels;
}

int PNGReader::getBufferSize() const {
    return width * height * channels;
}

Error PNGReader::readData(ColorChannel* buffer) {
    if(setjmp(png_jmpbuf(read))) {
        Error error = {"png file '"};
        error.message.append(path).append("' has used error callback");
        return error;
    }
    rowPointers = static_cast<ColorChannel**>(
        png_malloc(read, height * sizeof(ColorChannel*)));
    for(int i = 0; i < height; i++) {
        rowPointers[i] = (buffer + i * width * channels);
    }
    png_set_rows(read, info, reinterpret_cast<png_bytepp>(rowPointers));
    png_read_image(read, reinterpret_cast<png_bytepp>(rowPointers));
    return {};
}